#!/usr/bin/env/python
# -*- coding: utf-8 -*-

""" Module for the command line interface to data-fetcher. """
import logging
import sys
from argparse import Namespace
from enum import Enum
from pathlib import Path

from datafetcher.project_fetcher_refactor import ParallelFetcher

from .locations_report import LocationsReport
from .utilities import get_arg_parser, get_capo_settings, FlexLogger

_APPLICATION_NAME = 'datafetcher'

class ReturnCode(Enum):
    # TODO: confirm that each of these is tested
    SUCCESS = {'code': 0, 'text' : 'success'}
    MISSING_PROFILE = {'code': 1, 'text' : 'no CAPO profile provided'}
    MISSING_SETTING = {'code': 2, 'text' : 'missing required setting'}
    LOCATOR_SERVICE_TIMEOUT = {'code': 3, 'text' : 'request to locator service timed out'}
    TOO_MANY_SERVICE_REDIRECTS = {'code': 4, 'text' : 'too many redirects on locator service'}
    CATASTROPHIC_REQUEST_TIMEOUT = {'code': 5, 'text' : 'catastrophic error on request service'}
    PRODUCT_LOCATOR_NOT_FOUND = {'code': 6, 'text' : 'product locator not found'}
    CANNOT_OPEN_LOCATION_FILE = {'code': 7, 'text' : 'not able to open specified location file'}
    NGAS_FETCH_ERROR = {'code': 8, 'text' : 'error fetching file from NGAS server'}
    SIZE_MISMATCH = {'code': 9, 'text' : 'retrieved file not expected size'}

class DataFetcherRefactor:
    '''
        TO EXECUTE ALL DF TESTS: from datafetcher/, just run pytest

        example command line that should work with the correct local profile
        and a current product locator (these are changed whenever the
        database is rebuilt:
        datafetcher --profile local --output-dir ~/Downloads/
        --product-locator \
            uid://evla/execblock/93e1c0cd-76de-4e65-a3f2-d5fe55f386d8 \
            --verbose

        local.properties must have:
        - edu.nrao.archive.workflow.config
            .StartupSettings.temporaryDataDirectory pointing to a locally
            writable temp dir, e.g., /var/tmp
        - edu.nrao.archive.workflow.config.DeliverySettings.hostname must point
            to local computer
        - execution_site must NOT be DSOC or NAASC

    '''

    def __init__(self, args: Namespace, settings: dict):
        self.usage = self._build_usage_message()
        if args is None or settings is None:
            self._exit_with_error(ReturnCode.MISSING_SETTING)
        self.args = args
        self.settings = settings
        self.verbose = self.args.verbose

        # required arguments
        self.output_dir = args.output_dir
        if self.output_dir is None:
            self._exit_with_error(ReturnCode.MISSING_SETTING)

        # except AttributeError:
        #     logging.error('output location is required')
        #     self._exit_with_error(ReturnCode.MISSING_SETTING)

        if not Path(self.output_dir).is_dir():
            logging.error(f'output location {self.output_dir} not found')
            self._exit_with_error(ReturnCode.MISSING_SETTING)

        self._LOG = FlexLogger(self.__class__.__name__,
                               self.output_dir, self.verbose)
        self.profile = args.profile
        if self.profile is None:
            self._exit_with_error(ReturnCode.MISSING_PROFILE)

        # optional arguments
        self.is_dry = args.dry_run
        self.force  = args.force
        self.verbose = args.verbose or False

        if args.location_file is not None:
            if args.product_locator is not None:
                self._LOG.error('required: location file OR product locator '
                                '-- not both')
                self._exit_with_error(ReturnCode.MISSING_SETTING)
            self.location_file = args.location_file
        elif args.product_locator is not None:
            self.product_locator = args.product_locator
        else:
            self._LOG.error('you must specify either a location file or a '
                            'product locator')
            self._exit_with_error(ReturnCode.MISSING_SETTING)

        try:
            self.locations_report = self._get_locations()
            self.servers_report = self.locations_report.servers_report
        except SystemExit as exc:
            self._LOG.error(f'{exc}')
            if args.location_file:
                self._exit_with_error(ReturnCode.CANNOT_OPEN_LOCATION_FILE)
            else:
                self._exit_with_error(ReturnCode.PRODUCT_LOCATOR_NOT_FOUND)

        except Exception as exc:
            self._LOG.error(
                f'>>> throwing unexpected {type(exc)} during init: {exc}')
            raise

    def _exit_with_error(self, return_code: ReturnCode) -> int:
        print(self.usage)
        sys.exit(return_code.value)


    def exit_code_to_return_code(self, exit_code: int) -> ReturnCode:
        for return_code in ReturnCode:
            if return_code.value['code'] == exit_code:
                return return_code
        raise ValueError(f'unexpected return code: {exit_code}')

    @staticmethod
    def _build_usage_message() -> str:
        usage_str = 'Usage:'
        usage_str += """\tdatafetcher [-h]
             (--product-locator PRODUCT_LOCATOR | --location-file LOCATION_FILE)
             [--dry-run] [--output-dir OUTPUT_DIR] [-v, --verbose]
             [--profile PROFILE]\n"""
        usage_str += "\n\t\tReturn codes:\n"
        for return_code in ReturnCode:
            usage_str += f"\n\t\t\t{return_code.value['code']}: " \
                         f"{return_code.value['text']}"
        return usage_str

    def run(self):
        """
        launch the fetcher
        :return:
        """

        capo_settings = get_capo_settings(self.profile)
        try:
            return ParallelFetcher(
                self.args, capo_settings, self._LOG,
                self.servers_report).run()
        except SystemExit as exc:
            self._LOG.error(f'{exc}')
            self._exit_with_error(self.exit_code_to_return_code(exc.code))
        except Exception as exc:
            self._LOG.error(
                f'>>> throwing unexpected exception during run: {exc}')
            raise

    def _get_locations(self):
        try:
            capo_settings = get_capo_settings(self.profile)
            return LocationsReport(self._LOG,
                                   self.args,
                                   capo_settings)
        except Exception as exc:
            self._LOG.error(f'{exc}')
            if hasattr(self, 'location_file'):
                self._exit_with_error(ReturnCode.CANNOT_OPEN_LOCATION_FILE)
            elif hasattr(self, 'product_locator'):
                self._exit_with_error(ReturnCode.PRODUCT_LOCATOR_NOT_FOUND)
            else:
                # should never happen; we've already checked
                self._exit_with_error(ReturnCode.MISSING_SETTING)


def main():
    ''' this will be executed when fetcher is launched
        from the command line
    '''
    # try:
    parser = get_arg_parser()
    args = parser.parse_args()
    settings = get_capo_settings(args.profile)
    datafetcher = DataFetcherRefactor(args, settings)
    return datafetcher.run()


if __name__ == '__main__':
    main()
