import os
import tempfile
from pathlib import Path

# TODO: run pylint (has to be done outside IJ);
#  fix issues found by code analysis and SonarLint.
import pytest
from datafetcher.datafetcher_refactor import DataFetcherRefactor, ReturnCode
from datafetcher.utilities import get_arg_parser, ProductLocatorLookup, \
    ExecutionSite, RetrievalMode
from pycapo import CapoConfig

from .testing_utils import get_test_capo_settings, get_metadata_db_settings, \
    TEST_PROFILE, get_mini_locations_file

_LOCATION_FILENAME = 'locations.json'
_ASDM_XML = 'ASDM.xml'
_EB_EXTERNAL_NAME = 'sysstartS.58955.83384832176'

@pytest.fixture(autouse=True, scope='function')
def make_tempdir() -> None:
    umask = os.umask(0o000)
    top_level = tempfile.mkdtemp(prefix='datafetcher_test_', dir='/var/tmp')
    os.umask(umask)
    yield top_level

@pytest.fixture(scope='module')
def capo_settings():
    ''' get the capo settings once for whole module '''
    def get_capo_settings() -> CapoConfig:
        return get_test_capo_settings()
    capo_settings = get_capo_settings()
    yield capo_settings

@pytest.fixture(scope='module')
def settings(capo_settings):
    ''' grabs all the settings we will need for the datafetcher:
        Capo, database, test data
    '''
    db_settings = get_metadata_db_settings(TEST_PROFILE)
    test_data = _initialize_test_data(db_settings)
    yield Settings(capo_settings, db_settings, test_data)

def launch_datafetcher(args, capo_settings):
    ''' invoke the DF with these args and return the return/error code '''
    namespace = get_arg_parser().parse_args(args)
    return DataFetcherRefactor(namespace, capo_settings).run()

def test_settings_setup(settings):
    assert settings.capo_settings is not None
    assert isinstance(settings.capo_settings, dict)
    assert settings.db_settings is not None
    assert isinstance(settings.db_settings, dict)
    assert settings.test_data is not None
    assert isinstance(settings.test_data, dict)

def _initialize_test_data(db_settings):
    ext_name = '13B-014.sb28862036.eb29155786.56782.5720116088'

    product_locator = ProductLocatorLookup(db_settings) \
        .look_up_locator_for_ext_name(ext_name)
    return {'external_name': ext_name,
               'product_locator': product_locator}

def test_usage_statement_makes_sense():
    usage = DataFetcherRefactor._build_usage_message()
    assert usage.startswith('Usage:')
    lines = usage.split('\n')
    assert len(lines) >= len(ReturnCode) + 1

def test_no_args_prints_usage_and_croaks():
    with pytest.raises(TypeError):
        df = DataFetcherRefactor()
        assert df.usage == DataFetcherRefactor._build_usage_message()

def test_omitted_capo_profile_returns_expected_code(make_tempdir, settings):
    args = ['--product-locator', 'not-even',
            '--output-dir', str(make_tempdir)]
    with pytest.raises(SystemExit) as exc:
        launch_datafetcher(args, settings.capo_settings)
    assert exc.value.code['code'] == 1

def test_omitted_output_dir_returns_expected_code(settings):
    args = ['--product-locator', 'scooby-doo']
    try:
        with pytest.raises(SystemExit) as exc:
            launch_datafetcher(args, settings.capo_settings)
        assert exc.value.code['code'] == 1
    finally:
        # this test results in a log file being written to the
        # current directory. nuke it.
        curdir = Path('.')
        log_files = [file for file in curdir.iterdir()
                     if str(file).endswith('.log')]
        deleted = [file.unlink() for file in log_files]
        assert len(deleted) == len(log_files)

def test_two_locator_args_returns_expected_code(make_tempdir, settings):
    args = ['--product-locator', 'a_locator',
            '--location-file', 'location.json'
            '--output-dir', str(make_tempdir)]
    with pytest.raises(SystemExit) as exc:
        launch_datafetcher(args, settings.capo_settings)
    assert exc.value.code == ReturnCode.MISSING_SETTING.value['code']

class MockServiceTimeoutReturn:
    ''' Simulates locator request service timeout '''

    @staticmethod
    def run():
        return ReturnCode.LOCATOR_SERVICE_TIMEOUT

# @pytest.mark.skip('timeout')
def test_locator_service_timeout_returns_expected_code(monkeypatch,
                                                       settings,
                                                       make_tempdir):

    def mock_run(*args, **kwargs):
        return MockServiceTimeoutReturn.run()

    args = ['--product-locator', settings.test_data['product_locator'],
            '--output-dir', str(make_tempdir),
            '--profile', TEST_PROFILE]
    monkeypatch.setattr(DataFetcherRefactor, "run", mock_run)
    return_code = launch_datafetcher(args, settings.capo_settings)
    assert return_code == ReturnCode.LOCATOR_SERVICE_TIMEOUT


class MockTooManyServiceRedirectsReturn:

    @staticmethod
    def run():
        return ReturnCode.TOO_MANY_SERVICE_REDIRECTS

def test_too_many_service_redirects_returns_expected_code(monkeypatch,
                                                          settings,
                                                          make_tempdir):

    def mock_run(*args, **kwargs):
        return MockTooManyServiceRedirectsReturn().run()

    args = ['--product-locator', settings.test_data['product_locator'],
            '--output-dir', str(make_tempdir),
            '--profile', TEST_PROFILE]
    monkeypatch.setattr(DataFetcherRefactor, "run", mock_run)
    return_code = launch_datafetcher(args, settings.capo_settings)
    assert return_code == ReturnCode.TOO_MANY_SERVICE_REDIRECTS


class MockCatastrophicServiceErrorReturn:

    @staticmethod
    def run():
        return ReturnCode.CATASTROPHIC_REQUEST_TIMEOUT

def test_catastrophic_service_error_returns_expected_code(monkeypatch,
                                                          settings,
                                                          make_tempdir):

    def mock_run(*args, **kwargs):
        return MockCatastrophicServiceErrorReturn().run()

    args = ['--product-locator', settings.test_data['product_locator'],
            '--output-dir', str(make_tempdir),
            '--profile', TEST_PROFILE]

    monkeypatch.setattr(DataFetcherRefactor, "run", mock_run)
    return_code = launch_datafetcher(args, settings.capo_settings)
    assert return_code == ReturnCode.CATASTROPHIC_REQUEST_TIMEOUT

def test_product_locator_not_found_returns_expected_code(make_tempdir,
                                                         settings):
    args = ['--product-locator', 'not_a_locator',
            '--output-dir', str(make_tempdir),
            '--profile', TEST_PROFILE]
    with pytest.raises(SystemExit) as exc:
        launch_datafetcher(args, settings.capo_settings)
    assert exc.value.code == ReturnCode.PRODUCT_LOCATOR_NOT_FOUND.value

def test_unable_to_open_location_file_returns_expected_code(make_tempdir, settings):
    args = ['--location-file', 'location.json',
            '--output-dir', str(make_tempdir),
            '--profile', TEST_PROFILE]
    with pytest.raises(SystemExit) as exc:
        launch_datafetcher(args, settings.capo_settings)
    assert exc.value.code == ReturnCode.CANNOT_OPEN_LOCATION_FILE.value


class MockNgasFetchError:
    ''' Simulates a problem fetching file(s) from NGAS server '''
    @staticmethod
    def run():
        return ReturnCode.NGAS_FETCH_ERROR

def test_error_fetching_file_from_ngas_returns_expected_code(monkeypatch,
                                                             settings,
                                                             make_tempdir):

    def mock_run(*args, **kwargs):
        return MockNgasFetchError().run()

    args = ['--product-locator', settings.test_data['product_locator'],
            '--output-dir', str(make_tempdir),
            '--profile', TEST_PROFILE]
    monkeypatch.setattr(DataFetcherRefactor, "run", mock_run)
    return_code = launch_datafetcher(args, settings.capo_settings)
    assert return_code == ReturnCode.NGAS_FETCH_ERROR


class MockSizeMismatchError:
    ''' Simulates a retrieved file of unexpected size '''
    @staticmethod
    def run():
        return ReturnCode.SIZE_MISMATCH

def test_unexpected_size_returns_expected_code(monkeypatch,
                                               settings,
                                               make_tempdir):

    def mock_run(*args, **kwargs):
        return MockSizeMismatchError().run()

    args = ['--product-locator', settings.test_data['product_locator'],
            '--output-dir', str(make_tempdir),
            '--profile', TEST_PROFILE]
    monkeypatch.setattr(DataFetcherRefactor, "run", mock_run)
    return_code = launch_datafetcher(args, settings.capo_settings)
    assert return_code == ReturnCode.SIZE_MISMATCH

def test_nothing_retrieved_if_dry_locator(make_tempdir, settings):
    ''' Simulates dry run with product locator '''

    args = ['--product-locator', settings.test_data['product_locator'],
            '--output-dir', str(make_tempdir),
            '--profile', TEST_PROFILE,
            '--dry-run'
            ]
    launch_datafetcher(args, settings.capo_settings)
    tempdir_files = Path(make_tempdir).iterdir()
    for file in tempdir_files:
        if not str(file).endswith('.log') \
                and not str(file).endswith('.json'):
            pytest.fail('dry run with product locator -DID- fetch files')

def test_nothing_retrieved_if_dry_file(make_tempdir, settings):
    '''
        dry run with locations file
    :return:

    '''

    locations_file = get_mini_locations_file(Path(make_tempdir, _LOCATION_FILENAME))
    args = ['--location-file', str(locations_file),
            '--output-dir', str(make_tempdir),
            '--profile', TEST_PROFILE,
            '--dry-run'
            ]
    launch_datafetcher(args, settings.capo_settings)
    output_dir = Path(make_tempdir)
    tempdir_files = output_dir.iterdir()
    for file in tempdir_files:
        if not str(file).endswith('.log') \
                and not str(file).endswith('.json'):
            pytest.fail('dry run with locations file -DID- fetch files')

# @pytest.mark.skip('test_force_flag_overwrites_existing_file()')
def test_force_flag_overwrites_existing_file(make_tempdir, settings):
    top_level = Path(make_tempdir)
    location_file = get_mini_locations_file(top_level / _LOCATION_FILENAME)
    dest_dir = Path(top_level, _EB_EXTERNAL_NAME)
    dest_dir.mkdir(parents=True, exist_ok=True)

    # make a fake file to be overwritten
    fake_file = dest_dir / _ASDM_XML
    with open(fake_file, 'w') as to_write:
        to_write.write('alas, my days are numbered')

    args = ['--location-file', str(location_file),
            '--profile', TEST_PROFILE,
            '--output-dir', str(top_level),
            '--force']
    cmd = 'datafetcher ' + ' '.join(args)
    try:
        launch_datafetcher(args, settings.capo_settings)
    except SystemExit as ex:
        pytest.fail(f'{ex}')
    except Exception as exc:
        pytest.fail(f'{exc}')

    sizes = dict()

    # go thru destination directory recursively
    # and get everybody's size
    for file in dest_dir.rglob('*'):
        sizes[str(file)] = file.stat().st_size
    assert 37 == len(sizes)
    fake_size = fake_file.stat().st_size
    assert 9339 == fake_size

def test_no_overwrite_without_force(make_tempdir, settings):
    top_level = Path(make_tempdir)
    location_file = get_mini_locations_file(top_level / _LOCATION_FILENAME)
    dest_dir = Path(top_level, _EB_EXTERNAL_NAME)
    dest_dir.mkdir(parents=True, exist_ok=True)

    # make a fake file to be overwritten
    fake_file = dest_dir / _ASDM_XML
    with open(fake_file, 'w') as to_write:
        to_write.write('alas, my days are numbered')

    args = ['--location-file', str(location_file),
            '--profile', TEST_PROFILE,
            '--output-dir', str(top_level)
            ]

    with pytest.raises(SystemExit) as exc:
        launch_datafetcher(args, settings.capo_settings)
    assert exc.value.code == ReturnCode.NGAS_FETCH_ERROR.value

    sizes = dict()
    for file in dest_dir.rglob('*'):
        sizes[str(file)] = file.stat().st_size
    assert len(sizes) < 37
    fake_size = fake_file.stat().st_size
    assert 26 == fake_size

def test_more_output_when_verbose(make_tempdir, settings):
    top_level = Path(make_tempdir)
    location_file = get_mini_locations_file(top_level / _LOCATION_FILENAME)

    args = ['--location-file', str(location_file),
            '--profile', TEST_PROFILE,
            '--output-dir', str(make_tempdir),
            '--verbose']

    return_code = launch_datafetcher(args, settings.capo_settings)
    assert return_code == ReturnCode.SUCCESS.value['code']

    num_files_expected = 37
    retrieved = [file for file in top_level.rglob('*')
                 if file.is_file()]
    assert num_files_expected == len(retrieved) - 2

    verbose_logfile = None
    for file in retrieved:
        if str(file).endswith('.log'):
            verbose_logfile = file
            break

    assert verbose_logfile is not None
    verbose_log_size = verbose_logfile.stat().st_size
    assert verbose_log_size > 0

    # get rid of all the files we downloaded, plus the log
    deleted = [file.unlink() for file in retrieved
               if not str(file).endswith('.json')]
    assert len(deleted) >= num_files_expected

    ### same download, but without verbose logging
    args = ['--location-file', str(location_file),
            '--profile', TEST_PROFILE,
            '--output-dir', str(make_tempdir)]

    return_code = launch_datafetcher(args, settings.capo_settings)
    assert return_code == ReturnCode.SUCCESS.value['code']

    retrieved = [file for file in top_level.rglob('*')]
    assert len(retrieved) == num_files_expected + 3

    logfile = None
    for file in retrieved:
        if str(file).endswith('.log'):
            logfile = file
            break
    assert logfile is not None

    logsize = logfile.stat().st_size
    # successful download, non-verbose logging,
    # should result in zero-size log file
    assert logsize == 0


def test_copy_attempt_throws_sys_exit_service_error(make_tempdir, settings):
    local_exec_site = settings.capo_settings['execution_site']

    # use site from non-local profile to guarantee copy attempt
    settings.capo_settings['execution_site'] = ExecutionSite.DSOC

    args = ['--product-locator', settings.test_data['product_locator'],
            '--output-dir', str(make_tempdir),
            '--profile', 'nmprod']
    namespace = get_arg_parser().parse_args(args)
    fetcher = DataFetcherRefactor(namespace, settings.capo_settings)

    servers_report = fetcher.servers_report
    for server in servers_report:
        entry = servers_report[server]
        assert entry['retrieve_method'].value == RetrievalMode.COPY.value

    # let's try just one file so we're not sitting here all day
    for server in servers_report:
        entry = servers_report[server]
        servers_report = {server: entry}
        break
    fetcher.servers_report = servers_report
    assert fetcher.servers_report[server] is not None
    files = fetcher.servers_report[server]['files']
    fetcher.servers_report[server]['files'] = [files[0]]

    try:
        with pytest.raises(SystemExit) as exc:
            fetcher.run()
        assert exc.value.code == \
               ReturnCode.CATASTROPHIC_REQUEST_TIMEOUT.value
    finally:
        settings.capo_settings['execution_site'] = local_exec_site


# TODO: the rest of these

# def test_missing_profile_argument_returns_expected_code():
#     raise NotImplementedError

# def test_missing_profile_value_returns_expected_code():
#     raise NotImplementedError

# def test_invalid_profile_returns_expected_code():
#     raise NotImplementedError
#
# def test_dies_with_bad_server_info():
#
# def test_sys_exit_file_error_on_bad_destination():
#     raise NotImplementedError

# def test_gets_vlbas_from_report_file():
#     raise NotImplementedError
#
# def test_gets_large_vla_ebs_from_report_file():
#     raise NotImplementedError
#
# def test_gets_images_from_report_file():
#     raise NotImplementedError
#
# def test_gets_calibration_from_report_file():
#     raise NotImplementedError
#
# def test_gets_calibration_from_locator():
#     raise NotImplementedError

class Settings:

    def __init__(self, capo_settings, db_settings, test_data):
        self.capo_settings = capo_settings
        self.db_settings = db_settings
        self.test_data = test_data
